import os

import click

from . import dev
from .backend import (
    build_containers,
    exec_container,
    get_container_status,
    run_container,
)


@dev.command()
@click.argument('container')
@click.option('--shell', default='sh')
@click.pass_obj
def enter(kubetools_config, container, shell):
    '''
    Enter a shell in a running container.
    '''

    click.echo('--> Entering {0}'.format(click.style(container, bold=True)))
    exec_container(kubetools_config, container, [shell])


@dev.command()
@click.argument('container')
@click.pass_obj
def attach(kubetools_config, container):
    '''
    Attach to the main process in a container.
    '''

    container_status = get_container_status(kubetools_config, container)

    click.echo('--> Attaching to {0}'.format(click.style(container, bold=True)))
    click.echo('    use ctrl + p, ctrl + q to escape')

    # We *don't* use run_process here because when you escape the process exits
    # with an error code (go docker).
    os.system('docker attach {0}'.format(container_status['id']))


@dev.command(name='exec')
@click.argument('container')
@click.argument(
    'command',
    nargs=-1,
    required=True,
)
@click.pass_obj
def exec_(kubetools_config, container, command):
    '''
    Run a command in an existing container.
    '''

    click.echo('--> Building any out of date containers')
    build_containers(kubetools_config, [container])
    click.echo()

    click.echo('--> Executing in {0}: {1}'.format(container, command))
    return exec_container(kubetools_config, container, command)


@dev.command()
@click.argument('container')
@click.argument(
    'command',
    nargs=-1,
    required=True,
)
@click.option(
    'envvars', '-e', '--envvar',
    '--envar',  # legacy support TODO: remove!
    multiple=True,
    help='Environment variables to pass into the container.',
)
@click.pass_obj
def run(kubetools_config, container, command, envvars=None):
    '''
    Run a command in a new container.
    '''

    click.echo('--> Building any out of date containers')
    build_containers(kubetools_config, [container])
    click.echo()

    click.echo('--> Running in {0}: {1}'.format(container, command))
    return run_container(kubetools_config, container, command, envvars=envvars)



MANAGED_BY_ANNOTATION_KEY = 'app.kubernetes.io/managed-by'

GIT_BRANCH_ANNOTATION_KEY = 'kubetools/git_branch'
GIT_TAG_ANNOTATION_KEY = 'kubetools/git_tag'
GIT_COMMIT_ANNOTATION_KEY = 'kubetools/git_commit'

PROJECT_NAME_LABEL_KEY = 'kubetools/project_name'
ROLE_LABEL_KEY = 'kubetools/role'
NAME_LABEL_KEY = 'kubetools/name'



from kubetools.cli.git_utils import get_git_info
from kubetools.config import load_kubetools_config
from kubetools.constants import (
    ROLE_LABEL_KEY,
)
from kubetools.deploy.image import ensure_docker_images
from kubetools.deploy.util import log_actions
from kubetools.kubernetes.api import (
    create_deployment,
    create_job,
    create_namespace,
    create_service,
    delete_job,
    deployment_exists,
    get_object_name,
    list_deployments,
    list_namespaces,
    list_services,
    namespace_exists,
    service_exists,
    update_deployment,
    update_namespace,
    update_service,
)
from kubetools.kubernetes.config import (
    generate_kubernetes_configs_for_project,
    generate_namespace_config,
)


# Deploy/upgrade
# Handles deploying new services and upgrading existing ones

def get_deploy_objects(
    build,
    app_dirs,
    replicas=None,
    default_registry=None,
    extra_envvars=None,
    extra_annotations=None,
    ignore_git_changes=False,
    custom_config_file=False,
):
    all_services = []
    all_deployments = []
    all_jobs = []

    envvars = {
        'KUBE_ENV': build.env,
        'KUBE_NAMESPACE': build.namespace,
    }
    if extra_envvars:
        envvars.update(extra_envvars)

    annotations = {
        'kubetools/env': build.env,
        'kubetools/namespace': build.namespace,
    }
    if extra_annotations:
        annotations.update(extra_annotations)

    namespace = generate_namespace_config(build.namespace, base_annotations=annotations)

    for app_dir in app_dirs:
        commit_hash, git_annotations = get_git_info(app_dir, ignore_git_changes)
        annotations.update(git_annotations)

        kubetools_config = load_kubetools_config(
            app_dir,
            env=build.env,
            namespace=build.namespace,
            app_name=app_dir,
            custom_config_file=custom_config_file,
        )

        context_to_image = ensure_docker_images(
            kubetools_config, build, app_dir,
            commit_hash=commit_hash,
            default_registry=default_registry,
        )

        services, deployments, jobs = generate_kubernetes_configs_for_project(
            kubetools_config,
            envvars=envvars,
            context_name_to_image=context_to_image,
            base_annotations=annotations,
            replicas=replicas or 1,
        )

        all_services.extend(services)
        all_deployments.extend(deployments)
        all_jobs.extend(jobs)

    existing_deployments = {
        get_object_name(deployment): deployment
        for deployment in list_deployments(build.env, build.namespace)
    }

    # If we haven't been provided an explicit number of replicas, default to using
    # anything that exists live when available.
    if replicas is None:
        for deployment in all_deployments:
            existing_deployment = existing_deployments.get(get_object_name(deployment))
            if existing_deployment:
                deployment['spec']['replicas'] = existing_deployment.spec.replicas

    return namespace, all_services, all_deployments, all_jobs


def log_deploy_changes(
    build, namespace, services, deployments, jobs,
    message='Executing changes:',
    name_formatter=lambda name: name,
):
    existing_namespace_names = set(
        get_object_name(namespace)
        for namespace in list_namespaces(build.env)
    )
    existing_service_names = set(
        get_object_name(service)
        for service in list_services(build.env, build.namespace)
    )
    existing_deployment_names = set(
        get_object_name(deployment)
        for deployment in list_deployments(build.env, build.namespace)
    )

    deploy_service_names = set(
        get_object_name(service) for service in services
    )
    deploy_deployment_names = set(
        get_object_name(deployment) for deployment in deployments
    )
    deploy_namespace_name = set((build.namespace,))

    new_namespace = deploy_namespace_name - existing_namespace_names

    new_services = deploy_service_names - existing_service_names
    update_services = deploy_service_names - new_services

    new_deployments = deploy_deployment_names - existing_deployment_names
    update_deployments = deploy_deployment_names - new_deployments

    with build.stage(message):
        log_actions(build, 'CREATE', 'namespace', new_namespace, name_formatter)
        log_actions(build, 'CREATE', 'service', new_services, name_formatter)
        log_actions(build, 'CREATE', 'deployment', new_deployments, name_formatter)
        log_actions(build, 'UPDATE', 'service', update_services, name_formatter)
        log_actions(build, 'UPDATE', 'deployment', update_deployments, name_formatter)


def execute_deploy(build, namespace, services, deployments, jobs, delete_completed_jobs=True):
    # Split services + deployments into app (main) and dependencies
    depend_services = []
    main_services = []

    for service in services:
        if service['metadata']['labels'][ROLE_LABEL_KEY] == 'app':
            main_services.append(service)
        else:
            depend_services.append(service)

    depend_deployments = []
    main_deployments = []
    for deployment in deployments:
        if deployment['metadata']['labels'][ROLE_LABEL_KEY] == 'app':
            main_deployments.append(deployment)
        else:
            depend_deployments.append(deployment)

    # Now execute the deploy process
    if namespace:
        with build.stage('Create and/or update namespace'):
            if namespace_exists(build.env, namespace):
                build.log_info(f'Update namespace: {get_object_name(namespace)}')
                update_namespace(build.env, namespace)
            else:
                build.log_info(f'Create namespace: {get_object_name(namespace)}')
                create_namespace(build.env, namespace)

    if depend_services:
        with build.stage('Create and/or update dependency services'):
            for service in depend_services:
                if service_exists(build.env, build.namespace, service):
                    build.log_info(f'Update service: {get_object_name(service)}')
                    update_service(build.env, build.namespace, service)
                else:
                    build.log_info(f'Create service: {get_object_name(service)}')
                    create_service(build.env, build.namespace, service)

    if depend_deployments:
        with build.stage('Create and/or update dependency deployments'):
            for deployment in depend_deployments:
                if deployment_exists(build.env, build.namespace, deployment):
                    build.log_info(f'Update deployment: {get_object_name(deployment)}')
                    update_deployment(build.env, build.namespace, deployment)
                else:
                    build.log_info(f'Create deployment: {get_object_name(deployment)}')
                    create_deployment(build.env, build.namespace, deployment)

    noexist_main_services = []
    exist_main_services = []
    for service in main_services:
        if not service_exists(build.env, build.namespace, service):
            noexist_main_services.append(service)
        else:
            exist_main_services.append(service)

    if noexist_main_services:
        with build.stage('Create any app services that do not exist'):
            for service in noexist_main_services:
                build.log_info(f'Create service: {get_object_name(service)}')
                create_service(build.env, build.namespace, service)

    noexist_main_deployments = []
    exist_main_deployments = []
    for deployment in main_deployments:
        if not deployment_exists(build.env, build.namespace, deployment):
            noexist_main_deployments.append(deployment)
        else:
            exist_main_deployments.append(deployment)

    if noexist_main_deployments:
        with build.stage('Create any app deployments that do not exist'):
            for deployment in noexist_main_deployments:
                build.log_info(f'Create deployment: {get_object_name(deployment)}')
                create_deployment(build.env, build.namespace, deployment)

    if jobs:
        with build.stage('Execute upgrades'):
            for job in jobs:
                build.log_info(f'Create job: {get_object_name(job)}')
                create_job(build.env, build.namespace, job)
                if delete_completed_jobs:
                    delete_job(build.env, build.namespace, job)

    if exist_main_deployments:
        with build.stage('Update existing app deployments'):
            for deployment in exist_main_deployments:
                build.log_info(f'Update deployment: {get_object_name(deployment)}')
                update_deployment(build.env, build.namespace, deployment)

    if exist_main_services:
        with build.stage('Update existing app services'):
            for service in exist_main_services:
                build.log_info(f'Update service: {get_object_name(service)}')
                update_service(build.env, build.namespace, service)



import click

from kubetools.exceptions import KubeDevError
from kubetools.settings import get_settings

from . import dev
from .backend import (
    build_containers,
    destroy_containers,
    find_container_for_config,
    get_containers_status,
    print_containers,
    run_container,
    start_containers,
    stop_containers,
    up_containers,
)


@dev.command()
@click.option(
    '--active-only', '--active',
    is_flag=True,
    help='Only show containers from the active environment (--env).',
)
@click.pass_obj
def status(kubetools_config, active_only):
    '''
    Display info on the local dev environment.
    '''

    click.echo('--> Loading environments...')
    print_containers(kubetools_config, all_environments=not active_only)


@dev.command()
@click.argument('containers', nargs=-1)
@click.pass_obj
def destroy(kubetools_config, containers=None):
    '''
    Stop and remove containers.
    '''

    click.echo('--> Destroying containers')
    destroy_containers(kubetools_config, containers)
    click.echo('--> Dev environment down')


@dev.command()
@click.argument('containers', nargs=-1)
@click.pass_obj
def start(kubetools_config, containers=None):
    '''
    Start stopped containers.
    '''

    click.echo('--> Starting containers')
    start_containers(kubetools_config, containers)
    click.echo('--> Containers started')
    click.echo()

    # Always print all containers
    print_containers(kubetools_config)


@dev.command()
@click.argument('containers', nargs=-1)
@click.pass_obj
def stop(kubetools_config, containers=None):
    '''
    Stop running containers.
    '''

    click.echo('--> Stopping containers')
    stop_containers(kubetools_config, containers)
    click.echo('--> Containers stopped')
    click.echo()

    # Always print all containers
    print_containers(kubetools_config)


@dev.command()
@click.argument('containers', nargs=-1)
@click.pass_context
def restart(ctx, containers):
    '''
    Restart containers.
    '''

    ctx.invoke(stop, containers=containers)
    ctx.invoke(start, containers=containers)


@dev.command()
@click.argument('containers', nargs=-1)
@click.option('destroy_containers', '--destroy', is_flag=True)
@click.option('--no-upgrade', is_flag=True)
@click.pass_obj
@click.pass_context
def reload(
    ctx, kubetools_config,
    containers=None,
    destroy_containers=False,
    no_upgrade=False,
):
    '''
    Reload the dev environment.
    '''

    if destroy_containers:
        ctx.invoke(destroy, containers=containers)
    else:
        ctx.invoke(stop, containers=containers)

    click.echo()

    # Then up with all the arguments/options
    ctx.invoke(up, containers=containers, no_upgrade=no_upgrade)


@dev.command()
@click.argument('containers', nargs=-1)
@click.option(
    '--no-upgrade',
    is_flag=True,
    help='Disable running of any upgrades.',
)
@click.pass_obj
def up(
    kubetools_config,
    containers=None,
    no_upgrade=False,
    is_testing=False,
):
    '''
    Create and/or start containers and run any upgrades.
    '''

    container_statuses = get_containers_status(kubetools_config)

    # Figure out containers, if any, to upgrade
    upgrade_containers = []
    if not no_upgrade:
        if containers:
            upgrade_containers = containers
        else:
            upgrade_containers = [
                container_name
                for container_name, container in container_statuses.items()
            ]

    # Build the container(s)
    click.echo('--> Building any out of date containers')
    build_containers(kubetools_config, containers)
    click.echo()

    # Up the container(s)
    click.echo('--> Starting containers')
    up_containers(kubetools_config, containers)

    click.echo('--> Dev environment up')
    click.echo()

    if upgrade_containers:
        click.echo('--> Running upgrades')

        for upgrade in kubetools_config.get('upgrades', []):
            apply_when_testing = upgrade.get('conditions', {}).get('test', True)
            if is_testing and not apply_when_testing:
                continue

            container_name = find_container_for_config(
                kubetools_config, upgrade,
            )

            if containers and container_name not in upgrade_containers:
                continue

            click.echo('--> Running upgrade {0} in container {1}'.format(
                click.style(upgrade.get(
                    'name',
                    ' '.join(upgrade['command']),
                ), bold=True),
                container_name,
            ))

            run_container(
                kubetools_config,
                container_name,
                upgrade['command'],
            )
            click.echo()

    # Always print all containers
    print_containers(kubetools_config)
    click.echo()

    click.echo(''.join((
        click.style('Use `', 'blue'),
        click.style('ktd logs', bold=True),
        click.style('` to see what the containers are up to', 'blue'),
    )))
    click.echo(''.join((
        click.style('Use `', 'blue'),
        click.style('ktd attach <container>', bold=True),
        click.style('` to attach to a running container', 'blue'),
    )))


@dev.command()
@click.argument('arguments', nargs=-1)
@click.option(
    '--keep-containers',
    is_flag=True,
    help="Don't remove any test environment containers after completion",
)
@click.pass_obj
@click.pass_context
def test(ctx, kubetools_config, keep_containers=False, arguments=None):
    '''
    Execute tests in a new environment.

    You can pass extra arguments to the test command like so:

    ktd test -- --ipdb-failures
    '''

    DEV_DEFAULT_ENV = get_settings().DEV_DEFAULT_ENV

    # Set the env to test if not already overridden
    if kubetools_config.get('env', DEV_DEFAULT_ENV) == DEV_DEFAULT_ENV:
        kubetools_config['env'] = 'test'

    try:
        ctx.invoke(up, is_testing=True)

        click.echo()
        click.echo('--> Executing tests...')

        tests = kubetools_config.get('tests')

        if not tests:
            raise KubeDevError('No tests provided in kubetools config!')

        for test in tests:
            command = test['command']

            if arguments:
                command.extend(arguments)

            container_name = find_container_for_config(
                kubetools_config, test,
            )

            click.echo('--> Running test {0} in container {1}'.format(
                click.style(test.get(
                    'name',
                    ' '.join(command),
                ), bold=True),
                container_name,
            ))

            run_container(
                kubetools_config,
                container_name,
                command,
                envvars=test.get('environment', []),
            )
            click.echo()

    except Exception:
        click.echo(click.style('Exception!', 'red', bold=True))
        raise

    else:
        click.echo(click.style('--> Tests complete', 'green'))

    finally:
        if not keep_containers:
            ctx.invoke(destroy)



from functools import lru_cache

import docker
import requests

from kubetools.dev.process_util import run_process
from kubetools.exceptions import KubeDevError
from kubetools.log import logger

from .config import (
    create_compose_config,
    dockerise_label,
    get_all_containers,
    get_compose_filename,
    get_compose_name,
)


@lru_cache(maxsize=1)
def get_docker_client():
    client = docker.from_env()

    try:
        client.ping()

    except (docker.errors.APIError, requests.exceptions.ConnectionError) as e:
        raise KubeDevError((
            'Could not connect to Docker, is it running?\n'
            'Error: {0}'
        ).format(e))

    return client


def ensure_docker_dev_network():
    '''
    Ensure we have a dev Docker network for all our containers to belong to.
    '''

    docker_client = get_docker_client()

    network_names = [
        network.name
        for network in docker_client.networks.list()
    ]

    if 'dev' not in network_names:
        docker_client.networks.create('dev')


def get_all_docker_dev_network_containers():
    '''
    Gets the container names for everything using the global dev network.

    Note only containers created with the default "dev" env become part of this
    network - any custom env (eg when testing or `--env`) won't be included.
    '''

    docker_client = get_docker_client()

    labels = [
        # Ignore containers created by run/exec/etc (not services)
        'com.docker.compose.oneoff=False',
        # Load anything from a compose project
        'com.docker.compose.project',
    ]

    docker_containers = docker_client.containers.list(all=True, filters={
        'label': labels,
    })

    return docker_containers


def get_containers_status(
    kubetools_config,
    container_name=None,
    all_environments=False,
):
    '''
    Get the status of any containers for the current Kubetools project.
    '''

    docker_client = get_docker_client()

    labels = [
        # Ignore containers created by run/exec/etc (not services)
        'com.docker.compose.oneoff=False',
    ]

    if all_environments:
        # Filter by *any* docker compose - we filter out containers that don't
        # belong to this project below - this maintains compatability with
        # kubetools <8 that didn't write the kubetools.project.env label.
        labels.append('com.docker.compose.project')
    else:
        # Filter by the project/env name to only fetch this envs containers
        labels.append('com.docker.compose.project={0}'.format(
            get_compose_name(kubetools_config),
        ))

    if container_name:
        labels.append(
            'com.docker.compose.service={0}'.format(container_name),
        )

    logger.debug('Listing Docker containers with labels={0}'.format(labels))
    docker_containers = docker_client.containers.list(all=True, filters={
        'label': labels,
    })

    env_to_containers = {}
    docker_name = dockerise_label(kubetools_config['name'])

    for container in docker_containers:
        compose_project = container.labels['com.docker.compose.project']
        if all_environments:
            if not compose_project.startswith(docker_name):
                continue

            # For old Kubetools versions (<8) this label won't exist
            kubetools_name = container.labels.get('kubetools.project.name')
            if kubetools_name and kubetools_config['name'] != kubetools_name:
                continue

        env = container.labels.get('kubetools.project.env')
        # Compatability for existing containers created with kubetools <8
        if not env:
            env = compose_project.replace(docker_name, '')

        # Where the name is compose-name_container_N, get container
        name = container.name.split('_')[1]

        status = container.status == 'running'
        ports = []

        if container.attrs['NetworkSettings']['Ports']:
            for (
                local_port, host_port,
            ) in container.attrs['NetworkSettings']['Ports'].items():
                if not host_port:
                    continue

                ports.append({
                    'local': local_port,
                    'host': host_port[0]['HostPort'],
                })

        container_data = {
            'up': status,
            'ports': ports,
            'id': container.id,
            'labels': container.labels,
        }

        env_containers = env_to_containers.setdefault(env, {})

        if name in env_containers:
            raise ValueError((
                'Duplicate container for env {0}!: {1}({2})'
            ).format(env, name, container_data))

        env_containers[name] = container_data

    # Always provide the current environment, even if it's empty
    current_env = kubetools_config['env']
    if current_env not in env_to_containers:
        env_to_containers[kubetools_config['env']] = {}

    for env, containers in env_to_containers.items():
        for name, container in get_all_containers(kubetools_config):
            if name not in containers:
                containers[name] = {
                    'up': None,
                    'id': None,
                    'ports': [],
                }

            containers[name]['is_dependency'] = container.get('is_dependency', False)
            containers[name]['is_deployment'] = container.get('is_deployment', False)

    if all_environments:
        return env_to_containers
    return env_to_containers.get(kubetools_config['env'], {})


def get_container_status(kubetools_config, name):
    containers = get_containers_status(kubetools_config, container_name=name)
    return containers.get(name)


def run_compose_process(kubetools_config, command_args, **kwargs):
    # Ensure we have a compose file for this config
    create_compose_config(kubetools_config)

    compose_command = [
        'docker-compose',
        # Force us to look at the current directory, not relative to the compose
        # filename (ie .kubetools/compose-name.yml).
        '--project-directory', '.',
        # Name of the project (for the com.docker.compose.projectname label)
        '--project-name', get_compose_name(kubetools_config),
        # Filename of the YAML file to load
        '--file', get_compose_filename(kubetools_config),
    ]
    compose_command.extend(command_args)

    return run_process(compose_command, **kwargs)



import os

from time import sleep

import click
import requests

from pyretry import retry

from kubetools.exceptions import KubeDevError
from kubetools.log import logger
from kubetools.settings import get_settings

from .config import (
    get_all_containers,
    get_all_containers_by_name,
)
from .docker_util import (
    ensure_docker_dev_network,
    get_container_status,
    get_containers_status,
    run_compose_process,
)


def init_backend():
    ensure_docker_dev_network()


def http_get(url, timeout):
    try:
        response = requests.get(url, timeout=timeout)
        response.raise_for_status()

    except requests.RequestException as e:
        raise KubeDevError('Container start failed HTTP check: {0}'.format(e))


def find_container_for_config(kubetools_config, config):
    container_context = config['containerContext']
    all_containers = get_all_containers(kubetools_config)

    for name, container in all_containers:
        if container.get('containerContext') == container_context:
            return name
    else:
        raise KubeDevError((
            'No container found using containerContext: {0}'
        ).format(container_context))


def _build_container(kubetools_config, name, dockerfile):
    container_status = get_containers_status(kubetools_config)

    # Skip if the container is already up
    if name in container_status and container_status[name]['up']:
        return

    run_compose_process(
        kubetools_config,
        ('build', '--pull', name),
        hide_output=True,
    )


def build_containers(kubetools_config, names=None):
    all_containers = get_all_containers(kubetools_config)

    # Filter list of containers to build if specified
    if names:
        all_containers = [
            (name, container)
            for name, container in all_containers
            if name in names
        ]

    seen_dockerfiles = set()

    for name, config in all_containers:
        # Skip any image based containers
        if 'image' in config:
            continue

        dockerfile = config['build']['dockerfile']

        # Don't print the same `preBuildCommands` over and over!
        if dockerfile not in seen_dockerfiles:
            click.echo('--> Updating {0}'.format(dockerfile))

            pre_build_commands = config['build'].get('preBuildCommands')
            if pre_build_commands:
                click.echo('{0}: {1}:'.format(
                    click.style('Note', bold=True),
                    click.style((
                        'There are `preBuildCommands` specified for building '
                        '{0} which have not been executed'
                    ).format(dockerfile), 'yellow'),
                ))

                for command in pre_build_commands:
                    click.echo(' '.join(command))

        seen_dockerfiles.add(dockerfile)
        _build_container(kubetools_config, name, dockerfile)


def _up_container(kubetools_config, name):
    containers_status = get_containers_status(kubetools_config)

    # Skip if the container is already up
    if name in containers_status and containers_status[name]['up']:
        return

    # Up the container in the background
    run_compose_process(kubetools_config, ('up', '-d', name))

    # Get the new status of the container
    status = get_container_status(kubetools_config, name)

    if not status['up']:
        raise KubeDevError('Container {0} did not start properly'.format(
            click.style(name, bold=True),
        ))


def _probe_container(kubetools_config, name):
    settings = get_settings()

    containers = get_all_containers_by_name(kubetools_config)
    config = containers[name]

    status = get_container_status(kubetools_config, name)

    # Check for readinessProbe or probes (probes = readiness + liveness)
    probe = config.get('readinessProbe', config.get('probes'))
    if probe:
        timeout = probe.get('timeoutSeconds', 5)
        retries = probe.get('failureThreshold', 5)

        # Execute a command to check for container up?
        if 'exec' in probe:
            ready_command = probe['exec']['command']

            click.echo('--> Waiting for {0} to be ready with {1} (timeout={2}, retries={3})'.format(
                name, click.style(' '.join(ready_command), bold=True), timeout, retries,
            ))

            command = ['exec', '-T', name]
            command.extend(ready_command)

            run_with_retry = retry(
                KubeDevError,
                num_retries=retries,
                timeout=timeout,
            )(run_compose_process)

            run_with_retry(kubetools_config, command, hide_output=True)

        # Check HTTP status to check for container up?
        if 'httpGet' in probe:
            http_path = probe['httpGet'].get('path', '/')

            click.echo((
                '--> Waiting for {0} to be ready with HTTP GET '
                '{1} (timeout={2}, retries={3})'
            ).format(
                name, click.style(http_path, bold=True), timeout, retries,
            ))

            http_port = probe['httpGet'].get('port', 80)
            tcp_port = '{0}/tcp'.format(http_port)

            # Find the localhost port # matching the containers HTTP port
            target_port = None
            for port in status['ports']:
                if port['local'] == tcp_port:
                    target_port = port['host']
                    break

            http_url = 'http://{0}:{1}{2}'.format(
                settings.DEV_HOST, target_port, http_path,
            )

            get_with_retry = retry(
                KubeDevError,
                num_retries=retries,
                timeout=timeout,
            )(http_get)

            logger.debug('Executing HTTP check: {0}'.format(http_url))
            get_with_retry(http_url, timeout)

    # No probe? Sleep 2s and check it's still up as a super basic check
    else:
        sleep(2)

        # Get the new status of the container
        status = get_container_status(kubetools_config, name)

        if not status['up']:
            raise KubeDevError('Container {0} did not stay up for >2s'.format(
                click.style(name, bold=True),
            ))


def up_containers(kubetools_config, names=None):
    all_containers = get_all_containers(kubetools_config)

    # Filter matching containers if specified
    if names:
        all_containers = [
            (name, container)
            for name, container in all_containers
            if name in names
        ]

    # Up & probe all the containers in order
    for name, _ in all_containers:
        _up_container(kubetools_config, name)
        _probe_container(kubetools_config, name)


def destroy_containers(kubetools_config, names=None):
    containers_status = get_containers_status(kubetools_config)

    if containers_status:
        if not names:  # Shortcut: bring the entire environment down
            return run_compose_process(kubetools_config, ('down', '--remove-orphans'))

    all_containers = get_all_containers(kubetools_config)

    # Filter matching containers if specified
    if names:
        all_containers = [
            (name, container)
            for name, container in all_containers
            if name in names
        ]

    for name, config in all_containers:
        run_compose_process(kubetools_config, ('rm', '--stop', '--force', name))


def start_containers(kubetools_config, names=None):
    containers_status = get_containers_status(kubetools_config)

    # Check each container is stopped
    if names:
        for name in names:
            container_status = containers_status.get(name)

            if not container_status or container_status['up'] is None:
                raise KubeDevError((
                    'Container {0} does not exist and so cannot be started, '
                    'please create it with `ktd up {0}`.'.format(name)
                ))

    # Get the list of stopped containers
    else:
        names = [
            name
            for name, status in containers_status.items()
            # Check for both started & stopped - we only want to warn about
            # containers that don't exist (for ktd up).
            if status['up'] in (True, False)
        ]

        if not names:
            raise KubeDevError((
                'No stopped containers found, '
                'please use `ktd up` to create new containers.'
            ))

    run_compose_process(kubetools_config, ('start',) + tuple(names))


def stop_containers(kubetools_config, names=None):
    if not names:
        run_compose_process(kubetools_config, ('stop',))
        return

    containers_status = get_containers_status(kubetools_config)

    # Stop just these containers
    for name in names:
        container_status = containers_status.get(name)

        if container_status and container_status['up']:
            run_compose_process(kubetools_config, ('stop', name))


def run_container(kubetools_config, container, command, envvars=None):
    compose_command = ['run']

    if envvars:
        compose_command.extend(['-e{0}'.format(e) for e in envvars])

    compose_command.extend(["--entrypoint", " ".join(command)])
    compose_command.append(container)

    run_compose_process(kubetools_config, compose_command)


def exec_container(kubetools_config, container, command):
    compose_command = ['exec', container]
    compose_command.extend(command)

    run_compose_process(kubetools_config, compose_command)


def follow_logs(kubetools_config, containers, tail='all'):
    # Set this so we don't error and exit after 60s inactivity - which is the
    # ridiclous default value set by the docker-compose team.
    os.environ['COMPOSE_HTTP_TIMEOUT'] = '86400'

    args = ['logs', '--follow', '--tail={0}'.format(tail)]

    if containers:
        args.extend(containers)

    run_compose_process(kubetools_config, args)


def _print_containers(containers):
    container_infos = []
    settings = get_settings()

    for name, data in containers.items():
        port_strings = []

        for port in data['ports']:
            host_address = '{0}:{1}'.format(settings.DEV_HOST, port['host'])

            if port['local'] == '80/tcp':
                host_address = 'http://{0}'.format(host_address)

            port_strings.append(' / {0} -> {1}'.format(port['local'], host_address))

        link_text = ''.join(port_strings)

        if data['up'] is True:
            status_text = click.style('RUNNING', 'green')
        elif data['up'] is False:
            status_text = click.style('STOPPED', 'yellow')
        else:
            status_text = click.style('NOEXIST', 'red', bold=True)

        container_infos.append((
            name,
            '  - [{0}] {1}{2}{3}'.format(
                status_text,
                click.style(name, bold=True),
                ' (dependency)' if data.get('is_dependency') else '',
                link_text,
            ),
        ))

    # Sort alphabetically and print
    container_infos = sorted(
        container_infos,
        key=lambda c: '' if 'dependency' in c[1] else c[0],
    )
    for _, text in container_infos:
        click.echo(text)


def print_containers(kubetools_config, all_environments=False):
    envs_or_containers = get_containers_status(
        kubetools_config,
        all_environments=all_environments,
    )

    if all_environments:
        env_or_container_items = sorted(
            envs_or_containers.items(),
            key=lambda item: item[0] == kubetools_config['env'],
        )

        for i, (env, containers) in enumerate(env_or_container_items, 1):
            click.echo(click.style(
                '--> "{0}" environment state: (--env {0}{1})'.format(
                    env,
                    ', active' if env == kubetools_config['env'] else '',
                ),
                'blue',
            ))
            _print_containers(containers)

            if i < len(envs_or_containers):
                print()
    else:
        click.echo(click.style(
            '--> "dev" environment state:',
            'blue',
        ))
        _print_containers(envs_or_containers)
        containers = envs_or_containers

    # Print up/start messages depending on the container statuses
    up_statuses = [
        container['up']
        for container in containers.values()
    ]

    if None in up_statuses or False in up_statuses:
        click.echo()

    if None in up_statuses:
        click.echo(click.style(
            "Some containers don't exist, create them with: {0}".format(
                click.style('ktd up <container>', bold=True),
            ),
            'yellow',
        ))

    if False in up_statuses:
        click.echo(click.style(
            'Some containers are stopped; start them with: {0}'.format(
                click.style('ktd start <container>', bold=True),
            ),
            'yellow',
        ))



import json

from collections import defaultdict

import click
import yaml

from kubetools.config import load_kubetools_config
from kubetools.kubernetes.api import get_object_name
from kubetools.kubernetes.config import generate_kubernetes_configs_for_project

from . import cli_bootstrap


yaml.Dumper.ignore_aliases = lambda *args: True

FORMATTERS = {
    'json': lambda d: json.dumps(d, indent=4),
    'yaml': lambda d: yaml.dump(d),
}


@cli_bootstrap.command(help_priority=4)
@click.option(
    '--replicas',
    type=int,
    default=1,
    help='Default number of replicas for each app.',
)
@click.option(
    '-f', '--file',
    nargs=1,
    help='Specify a non-default Kubetools yml file to generate config from.',
    type=click.Path(exists=True),
)
@click.option(
    '--format', 'formatter',
    type=click.Choice(('json', 'yaml')),
    default='json',
    help='Specify the output format',
)
@click.argument(
    'app_dir',
    type=click.Path(exists=True, file_okay=False),
)
@click.pass_context
def config(ctx, replicas, file, app_dir, formatter):
    '''
    Generate and write out Kubernetes configs for a project.
    '''

    kubetools_config = load_kubetools_config(app_dir, custom_config_file=file)
    context_to_image = defaultdict(lambda: 'IMAGE')
    services, deployments, jobs = generate_kubernetes_configs_for_project(
        kubetools_config,
        replicas=replicas,
        context_name_to_image=context_to_image,
    )

    writer = FORMATTERS[formatter]

    for service in services:
        name = get_object_name(service)
        click.echo(f'Service: {click.style(name, bold=True)}')
        click.echo(writer(service))
        click.echo()

    for deployment in deployments:
        name = get_object_name(deployment)
        click.echo(f'Deployment: {click.style(name, bold=True)}')
        click.echo(writer(deployment))
        click.echo()

    for job in jobs:
        name = get_object_name(job)
        click.echo(f'Job: {click.style(name, bold=True)}')
        click.echo(writer(job))
        click.echo()



from configparser import ConfigParser
from functools import lru_cache
from os import access, listdir, path, X_OK

import click

from .log import logger


class KubetoolsSettings(object):
    KUBETOOLS_HOST = None
    KUBETOOLS_PORT = 80
    KUBETOOLS_API_VERSION = 4

    DEFAULT_KUBE_ENV = 'staging'  # default environment when speaking to the server

    DEV_DEFAULT_ENV = 'dev'  # default environment name in dev
    DEV_HOST = 'localhost'  # dev host to link people to (should map to 127.0.0.1)
    DEV_CONFIG_DIRNAME = '.kubetools'  # project directroy to generate compose config
    DEV_BACKEND = 'docker_compose'  # backend to use for development

    KUBETOOLS_SESSION = None  # user auth/session details (generated by the server)

    WAIT_SLEEP_TIME = 3
    WAIT_MAX_SLEEPS = 300 / WAIT_SLEEP_TIME

    def __init__(self, filename=None):
        self.filename = filename
        self.scripts = []


def get_settings_directory():
    return click.get_app_dir('kubetools', force_posix=True)


@lru_cache(maxsize=1)
def get_settings():
    settings_directory = get_settings_directory()
    settings_file = path.join(settings_directory, 'kubetools.conf')

    settings = KubetoolsSettings(filename=settings_file)

    if path.exists(settings_file):
        logger.info('Loading settings file: {0}'.format(settings_file))
        parser = ConfigParser()
        parser.read(settings_file)

        for option in parser.options('kubetools'):
            setattr(
                settings,
                option.upper().replace('-', '_'),
                parser.get('kubetools', option),
            )

    else:
        logger.info('No settings file: {0}'.format(settings_file))

    scripts_directory = path.join(settings_directory, 'scripts')
    if path.exists(scripts_directory):
        for filename in listdir(scripts_directory):
            if access(path.join(scripts_directory, filename), X_OK):
                settings.scripts.append(filename)

    return settings



from kubetools.deploy.util import delete_objects, get_app_objects, log_actions
from kubetools.kubernetes.api import (
    delete_deployment,
    delete_job,
    delete_service,
    list_deployments,
    list_jobs,
    list_services,
)


# Remove
# Handles removal of deployments, services and jobs in a namespace

def get_remove_objects(build, app_names=None, force=False):
    services_to_delete = get_app_objects(
        build,
        app_names,
        list_services,
        force=force,
    )
    deployments_to_delete = get_app_objects(
        build,
        app_names,
        list_deployments,
        force=force,
    )
    jobs_to_delete = get_app_objects(
        build,
        app_names,
        list_jobs,
        force=force,
    )

    return services_to_delete, deployments_to_delete, jobs_to_delete


def log_remove_changes(
    build, services, deployments, jobs,
    message='Executing changes:',
    name_formatter=lambda name: name,
):
    with build.stage(message):
        log_actions(build, 'DELETE', 'service', services, name_formatter)
        log_actions(build, 'DELETE', 'deployment', deployments, name_formatter)
        log_actions(build, 'DELETE', 'job', jobs, name_formatter)


def execute_remove(build, services, deployments, jobs):
    if services:
        with build.stage('Delete services'):
            delete_objects(build, services, delete_service)

    if deployments:
        with build.stage('Delete deployments'):
            delete_objects(build, deployments, delete_deployment)

    if jobs:
        with build.stage('Delete jobs'):
            delete_objects(build, jobs, delete_job)



#!/usr/bin/env python

import sys

import click

from .exceptions import KubeDevCommandError, KubeDevError, KubeError
from .log import logger
from .settings import get_settings


def run_cli(func):
    try:
        func()

    except KubeDevCommandError as e:
        message, stdout = e.args

        click.echo('--> {0} {1}'.format(
            click.style('Kubetools dev exception:', 'red', bold=True),
            message,
        ))
        click.echo(stdout)
        sys.exit(1)

    except KubeDevError as e:
        click.echo('--> {0} {1}'.format(
            click.style('Kubetools dev exception:', 'red', bold=True),
            e,
        ))
        sys.exit(1)

    except KubeError as e:
        if e.type == 'auth':
            settings = get_settings()

            logger.warning((
                'You are not logged into Kubetools!\n'
                'You need to login to your server, download the config '
                'from the top right and save it to:\n{0}\n'
            ).format(settings.filename))

        click.echo('--> {0} {1}'.format(
            click.style('Kubetools {0} exception:'.format(e.type), 'red', bold=True),
            e,
        ))
        sys.exit(1)

    except KeyboardInterrupt:
        click.echo()
        click.echo('Exiting on user request...')

    except Exception as e:
        click.echo('--> Unexpected exception: {0}'.format(
            click.style(
                '{0}{1}'.format(e.__class__.__name__, e.args),
                'red',
                bold=True,
            ),
        ))

        raise



'''
Shared Python module for handling Kubetools (kubetools.yml) config. Used by both
the dev client (ktd) and the server (lists KubetoolsClient as a requirement).
'''

from os import getcwd, path

import yaml

from pkg_resources import parse_version

from . import __version__
from .exceptions import KubeConfigError


# Config keys that can be filtered with a conditions: object
TOP_LEVEL_CONDITION_KEYS = (
    'upgrades',
    'deployments',
    'dependencies',
)

# Config keys that can have re-usable containerContext: names
TOP_LEVEL_CONTAINER_KEYS = TOP_LEVEL_CONDITION_KEYS + (
    'tests',
)


def load_kubetools_config(
    directory=None,
    app_name=None,
    # Filters for config items
    env=None,
    namespace=None,
    dev=False,  # when true disables env/namespace filtering (dev *only*)
    custom_config_file=False,
):
    '''
    Load Kubetools config files.

    Filtering:
        Most config items (deployments, dependencies, upgrades) can have conditions
        attached to them (eg dev: true). If an item has conditions, *either* dev or
        both env/namespace must match.

    Args:
        directory (str): directory to load ther config from (defaults to cwd)
        app_name (str): name of the app we're trying to load
        env (str): which envrionment to filter the config items by
        namespace (str): which namespace to filter the config items by
        dev (bool): filter config items by dev mode
    '''

    if custom_config_file:
        possible_filenames = (custom_config_file,)

    else:
        possible_filenames = (
            'kubetools.yml',
            'kubetools.yaml',
        )

    if directory:
        possible_files = [
            path.join(directory, filename)
            for filename in possible_filenames
        ]
    else:
        directory = getcwd()
        possible_files = []

        # Attempt parent directories back up to root
        while True:
            possible_files.extend([
                path.join(directory, filename)
                for filename in possible_filenames
            ])

            directory, splitdir = path.split(directory)
            if not splitdir:
                break

    config = None

    for filename in possible_files:
        try:
            with open(filename, 'r') as f:
                config = f.read()

        except IOError:
            pass
        else:
            break

    # If not present, this app deosn't support deploy/upgrade jobs (build/run only)
    if config is None:
        raise KubeConfigError((
            'Could not build app{0} as no kubetools config found!'
        ).format(' ({0})'.format(app_name) if app_name else ''))

    config = yaml.safe_load(config)
    config['_filename'] = filename

    # Check Kubetools version?
    if 'minKubetoolsVersion' in config:
        _check_min_version(config)

    # Apply an env name?
    if env:
        config['env'] = env

    # Filter out config items according to our conditions
    for key in TOP_LEVEL_CONDITION_KEYS:
        if key in config:
            config[key] = _filter_config_data(
                key, config[key],
                env=env,
                namespace=namespace,
                dev=dev,
            )

    # De-nest/apply any contextContexts
    for key in TOP_LEVEL_CONTAINER_KEYS:
        contexts = config.get('containerContexts', {})

        if key in config:
            config[key] = _expand_containers(
                key, config[key],
                contexts=contexts,
                dev=dev,
            )

    return config


def _check_min_version(config):
    running_version = parse_version(__version__)
    needed_version = parse_version(
        # Version must be a string
        str(config['minKubetoolsVersion']),
    )

    if needed_version > running_version:
        raise KubeConfigError(
            'Minimum Kubetools version not met, need {0} but got {1}'.format(
                needed_version, running_version,
            ),
        )


def _filter_config_data(key, items_or_object, env, namespace, dev):
    def is_match(item):
        return _conditions_match(
            item.get('conditions'),
            env=env,
            namespace=namespace,
            dev=dev,
        )

    if isinstance(items_or_object, list):
        return [
            item for item in items_or_object
            if is_match(item)
        ]

    elif isinstance(items_or_object, dict):
        return {
            key: item
            for key, item in items_or_object.items()
            if is_match(item)
        }

    else:
        raise KubeConfigError('Invalid type ({0}) for key: {1}'.format(
            type(items_or_object),
            key,
        ))


def _conditions_match(conditions, env, namespace, dev):
    # No conditions? We're good!
    if conditions is None:
        return True

    # Dev mode? Must have dev: true (or no conditions above)
    if dev:
        return conditions.get('dev') is True

    # If dev is set and nothing else, fail (dev only)!
    if conditions.get('dev') and len(conditions) == 1:
        return False

    # If we have envs but our env isn't present, fail!
    if 'envs' in conditions and env not in conditions['envs']:
        return False

    # We have namespaces but our namespace isn't present, fail!
    if 'namespaces' in conditions and namespace not in conditions['namespaces']:
        return False

    # If we have notNamespaces and our namespace is present, fail!
    if 'notNamespaces' in conditions and namespace in conditions['notNamespaces']:
        return False

    return True


def _expand_containers(key, items_or_object, contexts, dev):
    def do_expand(item):
        return _expand_container(
            item,
            contexts=contexts,
            dev=dev,
        )

    # List items have conditions and containerContext at the same level
    if isinstance(items_or_object, list):
        return [
            do_expand(item)
            for item in items_or_object
        ]

    # Named items have conditions top level. but containers are nested
    elif isinstance(items_or_object, dict):
        new_item = {}

        for key, item in items_or_object.items():
            if 'containers' in item:
                item['containers'] = {
                    k: do_expand(v)
                    for k, v in item.pop('containers').items()
                }

            new_item[key] = item

        return new_item

    else:
        raise KubeConfigError('Invalid type ({0}) for key: {1}'.format(
            type(items_or_object),
            key,
        ))


def _merge_config(base_config, new_config):
    for key, value in new_config.items():
        # If this key is a dict in the old config, merge those
        if key in base_config and isinstance(value, dict):
            _merge_config(base_config[key], new_config[key])
        else:
            base_config[key] = new_config[key]


def _expand_container(container, contexts, dev):
    # Expand any containerContext objects
    if 'containerContext' in container:
        context_name = container.get('containerContext')

        try:
            context = contexts[context_name]
        except KeyError:
            raise KubeConfigError('Missing containerContext: {0}'.format(
                context_name,
            ))

        # Merge in anything not explicitly defined on the container itself
        _merge_config(container, context)

    # Expand any dev keys from the dev config
    if 'dev' in container:
        dev_overrides = container.pop('dev')
        if dev:
            _merge_config(container, dev_overrides)

    # Apply any commandArguments
    if 'commandArguments' in container:
        # Duplicate the command here as it might be from a context so we don't
        # want to continuously append the arguments to the same base command.
        command = [cmd for cmd in container.pop('command')]
        command.extend(container.pop('commandArguments'))
        container['command'] = command

    return container



#!/usr/bin/env python

from kubetools.dev import dev
from kubetools.main import run_cli

# Import click command groups
from kubetools.dev import (  # noqa: F401, I100, I202
    container,
    environment,
    logs,
    scripts,
)


run_cli(dev)



from .util import make_dns_safe_name


def make_service_config(
    name, ports,
    node_port=True,
    labels=None,
    annotations=None,
):
    '''
    Builds a Kubernetes service configuration dict.
    '''

    labels = labels or {}
    annotations = annotations or {}

    # Build our ports list
    service_ports = []
    for port in ports:
        # Accept port dicts w/protocol/name/etc
        if isinstance(port, dict):
            # Default targetPort to same as port
            if 'targetPort' not in port:
                port['targetPort'] = port['port']

            service_ports.append(port)

        # And accept plain numbers
        else:
            service_ports.append({
                'port': port,
                'targetPort': port,
            })

    # The actual service spec
    service = {
        'apiVersion': 'v1',
        'kind': 'Service',
        'metadata': {
            'name': make_dns_safe_name(name),
            'labels': labels,
            'annotations': annotations,
        },
        'spec': {
            'selector': labels,
            'ports': service_ports,
        },
    }

    if node_port:
        service['spec']['type'] = 'NodePort'

    return service



from os import path
from unittest import mock, TestCase

import yaml

from kubetools.config import load_kubetools_config
from kubetools.kubernetes.api import get_object_name
from kubetools.kubernetes.config import generate_kubernetes_configs_for_project


def _assert_yaml_objects(objects, yaml_filename):
    with open(yaml_filename, 'r') as f:
        desired_objects = list(yaml.safe_load_all(f))

    objects.sort(key=get_object_name)
    desired_objects.sort(key=get_object_name)

    assert objects == desired_objects


def _test_configs(folder_name, **kwargs):
    app_dir = path.join('tests', 'configs', folder_name)

    kubetools_config = load_kubetools_config(app_dir, **kwargs)

    with mock.patch('kubetools.kubernetes.config.job.uuid4', lambda: 'UUID'):
        services, deployments, jobs = generate_kubernetes_configs_for_project(
            kubetools_config,
        )

    if services:
        _assert_yaml_objects(services, path.join(app_dir, 'k8s_services.yml'))
    if deployments:
        _assert_yaml_objects(deployments, path.join(app_dir, 'k8s_deployments.yml'))
    if jobs:
        _assert_yaml_objects(jobs, path.join(app_dir, 'k8s_jobs.yml'))


class TestKubernetesConfigGeneration(TestCase):
    def test_basic_app_configs(self):
        _test_configs('basic_app')

    def test_dependencies_configs(self):
        _test_configs('dependencies')

    def test_dev_overrides_configs(self):
        _test_configs('dev_overrides', dev=True)

    def test_k8s_container_passthrough_configs(self):
        _test_configs('k8s_container_passthrough')

    def test_multiple_deployments_configs(self):
        _test_configs('multiple_deployments')



import click

from tabulate import tabulate

from kubetools.constants import (
    GIT_BRANCH_ANNOTATION_KEY,
    GIT_COMMIT_ANNOTATION_KEY,
    GIT_TAG_ANNOTATION_KEY,
    NAME_LABEL_KEY,
    PROJECT_NAME_LABEL_KEY,
    ROLE_LABEL_KEY,
)
from kubetools.kubernetes.api import (
    get_object_annotations_dict,
    get_object_labels_dict,
    get_object_name,
    is_kubetools_object,
    list_deployments,
    list_jobs,
    list_replica_sets,
    list_services,
)

from . import cli_bootstrap


def _get_service_meta(service):
    meta_items = []
    for port in service.spec.ports:
        meta_items.append(f'port={port.port}, nodePort={port.node_port}')
    return ''.join(meta_items)


def _print_items(items, header_to_getter=None):
    header_to_getter = header_to_getter or {}

    headers = ['Name', 'Role', 'Project']
    headers.extend(header_to_getter.keys())
    headers = [click.style(header, bold=True) for header in headers]

    rows = []
    for item in items:
        labels = get_object_labels_dict(item)

        row = [
            get_object_name(item),
            labels.get(ROLE_LABEL_KEY),
        ]

        if not is_kubetools_object(item):
            row.append(click.style('NOT MANAGED BY KUBETOOLS', 'yellow'))
        else:
            row.append(labels.get(PROJECT_NAME_LABEL_KEY, 'unknown'))

        for getter in header_to_getter.values():
            row.append(getter(item))

        rows.append(row)

    click.echo(tabulate(rows, headers=headers, tablefmt='simple'))


def _get_node_ports(item):
    node_ports = []
    for port in item.spec.ports:
        if port.node_port:
            node_ports.append(f'{port.port}:{port.node_port}')
        else:
            node_ports.append(f'{port.port}')

    return ', '.join(node_ports)


def _get_ready_status(item):
    return f'{item.status.ready_replicas or 0}/{item.status.replicas}'


def _get_version_info(item):
    annotations = get_object_annotations_dict(item)
    bits = []
    for name, key in (
        ('branch', GIT_BRANCH_ANNOTATION_KEY),
        ('tag', GIT_TAG_ANNOTATION_KEY),
        ('commit', GIT_COMMIT_ANNOTATION_KEY),
    ):
        data = annotations.get(key)
        if data:
            bits.append(f'{name}={data}')

    return ', '.join(bits)


def _get_completion_status(item):
    return f'{item.status.succeeded}/{item.spec.completions}'


def _get_command(item):
    return get_object_annotations_dict(item).get('description')


@cli_bootstrap.command(help_priority=3)
@click.argument('namespace')
@click.argument('app', required=False)
@click.pass_context
def show(ctx, namespace, app):
    '''
    Show running apps in a given namespace.
    '''

    exists = False

    env = ctx.meta['kube_context']

    if app:
        click.echo(f'--> Filtering by app={app}')

    services = list_services(env, namespace)

    if services:
        exists = True

        if app:
            services = [s for s in services if get_object_name(s) == app]

        click.echo(f'--> {len(services)} Services')
        _print_items(services, {
            'Port(:nodePort)': _get_node_ports,
        })
        click.echo()

    deployments = list_deployments(env, namespace)

    if deployments:
        exists = True

        if app:
            deployments = [d for d in deployments if get_object_name(d) == app]

        click.echo(f'--> {len(deployments)} Deployments')

        _print_items(deployments, {
            'Ready': _get_ready_status,
            'Version': _get_version_info,
        })
        click.echo()

    if app:
        replica_sets = list_replica_sets(env, namespace)
        replica_sets = [
            r for r in replica_sets
            if r.metadata.labels.get(NAME_LABEL_KEY) == app
        ]

        click.echo(f'--> {len(replica_sets)} Replica sets')
        _print_items(replica_sets, {
            'Ready': _get_ready_status,
            'Version': _get_version_info,
        })
        click.echo()
    else:
        jobs = list_jobs(env, namespace)
        if jobs:
            exists = True

            click.echo(f'--> {len(jobs)} Jobs')
            _print_items(jobs, {
                'Completions': _get_completion_status,
                'Command': _get_command,
            })
            click.echo()

    if not exists:
        click.echo('Nothing to be found here 👀!')



import requests

from kubetools.exceptions import KubeBuildError
from kubetools.kubernetes.config import make_context_name

from .util import run_shell_command


def get_commit_hash_tag(context_name, commit_hash):
    '''
    Turn a commit hash into a Docker registry tag.
    '''

    return '-'.join((context_name, 'commit', commit_hash))


def get_docker_name(registry, app_name):
    return '{0}/{1}'.format(registry, app_name)


def get_docker_tag(registry, app_name, tag):
    # Tag the image like registry/app:tag
    docker_version = '{0}:{1}'.format(app_name, tag)
    # The full docker tag
    return '{0}/{1}'.format(registry, docker_version)


def get_docker_tag_for_commit(registry, app_name, context_name, commit_hash):
    return get_docker_tag(registry, app_name, get_commit_hash_tag(context_name, commit_hash))


def has_app_commit_image(registry, app_name, context_name, commit_hash):
    '''
    Check the registry has an app image for a certain commit hash.
    '''

    if registry is None:
        raise KubeBuildError(f'Invalid registry to build {context_name}: {registry}')

    commit_version = get_commit_hash_tag(context_name, commit_hash)
    url = 'http://{0}/v2/{1}/manifests/{2}'.format(registry, app_name, commit_version)

    response = requests.get(url)

    if response.status_code != 200:
        return False

    return True


def _get_container_contexts_from_config(app_config):
    context_name_to_build = {
        key: context['build']
        for key, context in app_config.get('containerContexts', {}).items()
        if 'build' in context
    }

    for deployment, data in app_config.get('deployments', {}).items():
        containers = data.get('containers')
        for name, container in containers.items():
            if 'containerContext' in container:
                continue

            if 'build' in container:
                context_name = make_context_name(deployment, name)
                if context_name in context_name_to_build:
                    raise KubeBuildError('Duplicate deployment/container')

                context_name_to_build[context_name] = container['build']

    return context_name_to_build


def ensure_docker_images(kubetools_config, build, *args, **kwargs):
    '''
    Ensures that our Docker registry has the specified image. If not we build
    and upload to the registry.
    '''

    project_name = kubetools_config['name']
    commit_hash = kwargs.get('commit_hash')

    with build.stage(f'Ensuring Docker images built for {project_name}={commit_hash}'):
        return _ensure_docker_images(kubetools_config, build, *args, **kwargs)


def _ensure_docker_images(
    kubetools_config, build, app_dir, commit_hash,
    default_registry=None,
    check_build_control=lambda build: None,
    additional_tags=None,
):
    if additional_tags is None:
        additional_tags = []
    project_name = kubetools_config['name']

    context_name_to_build = _get_container_contexts_from_config(kubetools_config)
    context_name_to_registry = {
        context_name: build_context.get('registry', default_registry)
        for context_name, build_context in context_name_to_build.items()
    }
    build_context_keys = list(context_name_to_build.keys())

    # Check if the image already exists in the registry
    if not build_context_keys or all(
        has_app_commit_image(
            context_name_to_registry[context_name],
            project_name,
            context_name,
            commit_hash,
        )
        for context_name in build_context_keys
    ):
        build.log_info((
            f'All Docker images for {project_name} commit {commit_hash} exists, '
            'skipping build'
        ))

        context_images = {
            # Build the context name -> image dict
            context_name: get_docker_tag_for_commit(
                context_name_to_registry[context_name],
                project_name,
                context_name,
                commit_hash,
            )
            for context_name in build_context_keys
        }

        return context_images

    # We're building something - let's find the previous commit we built
    commit_history = run_shell_command(
        'git', 'log', '--pretty=format:"%h"',
        cwd=app_dir,
    ).decode()

    commit_history = [
        commit.strip('"')
        for commit in commit_history.split()
    ]

    # Figure out the previous commit we built an image for
    previous_commit = None
    first_build_context = build_context_keys[0]
    first_build_registry = context_name_to_registry[first_build_context]
    for i, commit in enumerate(commit_history):
        if has_app_commit_image(
            first_build_registry,
            project_name,
            first_build_context,
            commit,
        ):
            previous_commit = commit
            break

        # We only search the most recent 100 commits before giving up, so as not
        # to overload the registry server.
        elif i >= 100:
            break

    # Check/abort as requested
    check_build_control(build)

    build.log_info(f'Building {project_name} @ commit {commit_hash}')

    # Now actually build the images
    context_images = {}

    for context_name, build_context in context_name_to_build.items():
        # Check/abort as requested
        check_build_control(build)

        registry = build_context.get('registry', default_registry)

        # Run pre docker commands?
        pre_build_commands = build_context.get('preBuildCommands', [])

        for command in pre_build_commands:
            # Check/abort as requested
            check_build_control(build)

            build.log_info(f'Executing pre-build command: {command}')

            # Run it, passing in the commit hashes as ENVars
            env = {
                'KUBE_ENV': build.env,
                'BUILD_COMMIT': commit_hash,
            }
            if previous_commit:
                env['PREVIOUS_BUILD_COMMIT'] = previous_commit

            run_shell_command(*command, cwd=app_dir, env=env)

        # The full docker tag
        docker_tag_for_commit = get_docker_tag_for_commit(
            registry,
            project_name,
            context_name,
            commit_hash,
        )
        additional_docker_tags = [
            get_docker_tag(registry, project_name, additional_tag)
            for additional_tag in additional_tags
        ]
        docker_tags = [docker_tag_for_commit]
        docker_tags.extend(additional_docker_tags)
        tag_arguments = []
        for docker_tag in docker_tags:
            tag_arguments.extend(['-t', docker_tag])

        # Build the image
        build.log_info((
            f'Building {project_name}/{context_name} '
            f'(file: {build_context["dockerfile"]}, commit: {commit_hash})'
        ))

        run_shell_command(
            'docker', 'build', '--pull',
            '-f', build_context['dockerfile'],
            *tag_arguments,
            '.',
            cwd=app_dir,
        )

        # Push the image and additional tags
        for docker_tag in docker_tags:
            build.log_info(f'Pushing docker image: {docker_tag}')
            run_shell_command('docker', 'push', docker_tag)

        context_images[context_name] = docker_tag_for_commit

    return context_images



import re

from collections import OrderedDict
from functools import lru_cache
from hashlib import md5
from os import makedirs, path

import click
import yaml

from kubetools.exceptions import KubeDevError
from kubetools.settings import get_settings

NON_COMPOSE_KEYS = (
    # Kubetools specific
    'minKubetoolsVersion',

    # Kubetools dev specific
    'devScripts',
    'containerContext',

    # Kubetools -> Kubernetes specific
    'servicePorts',
    'probes',
    'resources',

    # Kubernetes specific
    'livenessProbe',
    'readinessProbe',

    # Internal flags for `ktd status`
    'is_deployment',
    'is_dependency',
)

CONTAINER_KEYS = ('dependencies', 'deployments')
CONTAINER_KEY_TO_FLAG = {
    'dependencies': 'is_dependency',
    'deployments': 'is_deployment',
}


def dockerise_label(value):
    # Unfortunate egg from Docker engine pre label support, see:
    # https://github.com/docker/compose/issues/2119
    return re.sub(r'[^a-z0-9]', '', value.lower())


def get_project_name(kubetools_config):
    name = kubetools_config['name']
    env = kubetools_config['env']

    # Compose name is APP-ENV
    return '-'.join((name, env))


def get_compose_name(kubetools_config):
    name_env = get_project_name(kubetools_config)
    return dockerise_label(name_env)


def get_compose_dirname(kubetools_config):
    settings = get_settings()
    return path.join(
        path.dirname(kubetools_config['_filename']),
        settings.DEV_CONFIG_DIRNAME,
    )


def get_compose_filename(kubetools_config):
    env = kubetools_config['env']
    compose_filename = '{0}-compose.yml'.format(env)
    return path.join(get_compose_dirname(kubetools_config), compose_filename)


def get_all_containers(kubetools_config, container_keys=CONTAINER_KEYS):
    containers = []

    for key_name in container_keys:
        deployments = kubetools_config.get(key_name, {})

        for deployment_name, deployment_config in deployments.items():
            if 'containers' not in deployment_config:
                raise KubeDevError('Deployment {0} is missing containers'.format(
                    deployment_name,
                ))

            deployment_containers = deployment_config['containers']

            for container_name, config in deployment_containers.items():
                config[CONTAINER_KEY_TO_FLAG[key_name]] = True
                containers.append((container_name, config))

    return containers


def get_all_containers_by_name(kubetools_config, container_keys=CONTAINER_KEYS):
    return OrderedDict(get_all_containers(
        kubetools_config,
        container_keys=container_keys,
    ))


def _create_compose_service(kubetools_config, name, config, envvars=None):
    for invalid_build_key in (
        'preBuildCommands',
        'registry',
    ):
        if invalid_build_key in config.get('build', {}):
            config['build'].pop(invalid_build_key)

    # Because this is one of our containers (buildContexts are relevant to
    # the project) - setup a TTY and STDIN so we can attach and be interactive
    # when attached (eg ipdb).
    service = {
        'tty': True,
        'stdin_open': True,
        'labels': {
            'kubetools.project.name': kubetools_config['name'],
            'kubetools.project.env': kubetools_config['env'],
        },
    }

    service.update(config)

    # Translate k8s command/args to docker-compose entrypoint/command
    if 'command' in service:
        service['entrypoint'] = service.pop('command')
    if 'args' in config:
        service['command'] = service.pop('args')

    if 'build' in service and 'context' not in service['build']:
        service['build']['context'] = '.'

    if 'ports' in config:
        # Generate a consistent base port for this project/container combo
        hash_string = '{0}-{1}'.format(get_project_name(kubetools_config), name)

        # MD5 the string, integer that and then modulus 10k to shorten it down
        port_base = int(md5(hash_string.encode('utf-8')).hexdigest(), 16) % 10000
        # And bump by 10k so we don't stray into the privileged port range (<1025)
        port_base += 10000

        # Reassign ports with explicit host port numbers
        ports = []

        for port in config['ports']:
            if isinstance(port, dict):
                port = port['port']

            ports.append(
                '{0}:{1}'.format(int(port_base) + int(port), port),
            )

        service['ports'] = ports

    # Make our service - drop anything kubernetes or kubetools specific
    service = {
        key: value
        for key, value in service.items()
        if key not in NON_COMPOSE_KEYS
    }

    # Provide project-specific aliases for all containers. This means we can up
    # the same containers (eg mariadb x2) under the same dev network, but have
    # each app speak to it's own mariadb instance (eg app-mariadb).
    compose_name = kubetools_config['name']

    service['networks'] = {
        'default': {
            'aliases': [
                '{0}-{1}'.format(compose_name, name),
            ],
        },
    }

    # Add any *missing* extra envvars
    if envvars:
        environment = service.setdefault('environment', [])
        for envar in envvars:
            if envar not in environment:
                environment.append(envar)

    return service


@lru_cache(maxsize=1)
def get_dev_network_environment_variables():
    # This "fixes" a horrible circular dependency between config/docker_util
    from .docker_util import get_all_docker_dev_network_containers
    containers = get_all_docker_dev_network_containers()

    envvars = set()

    for container in containers:
        networks = container.attrs['NetworkSettings']['Networks']
        if 'dev' not in networks:
            continue

        aliases = networks['dev']['Aliases']
        for alias in aliases:
            if '-' in alias:
                break
        else:  # no alias with "-"
            continue

        envar = alias.upper().replace('-', '_')
        envvars.add('DEV_{0}={1}'.format(envar, alias))
    return list(envvars)


def create_compose_config(kubetools_config):
    # If we're not in a custom env, everything sits on the "dev" network. Envs
    # remain encapsulated inside their own network.
    DEV_DEFAULT_ENV = get_settings().DEV_DEFAULT_ENV
    ktd_env = kubetools_config.get('env', DEV_DEFAULT_ENV)
    dev_network = ktd_env == DEV_DEFAULT_ENV

    all_containers = get_all_containers(kubetools_config)

    envvars = [
        'KTD_ENV={0}'.format(ktd_env),
    ]

    dev_network_envvars = None
    if dev_network:
        dev_network_envvars = get_dev_network_environment_variables()
        if dev_network_envvars:
            envvars.extend(dev_network_envvars)

    click.echo('--> Injecting environment variables:')
    for envar in envvars:
        click.echo('    {0}'.format(envar))

    services = {
        name: _create_compose_service(
            kubetools_config, name, config,
            envvars=envvars,
        )
        for name, config in all_containers
    }

    compose_config = {
        'version': '3',
        'services': services,
    }

    if dev_network:
        compose_config['networks'] = {
            'default': {
                'external': {
                    'name': 'dev',
                },
            },
        }

    yaml_data = yaml.safe_dump(compose_config)

    compose_dirname = get_compose_dirname(kubetools_config)
    if not path.exists(compose_dirname):
        makedirs(compose_dirname)

    with click.open_file(get_compose_filename(kubetools_config), 'w') as f:
        f.write(yaml_data)



#!/usr/bin/env python

from kubetools.cli import cli_bootstrap
from kubetools.main import run_cli
# Import click command groups
from kubetools.cli import deploy, generate_config, push_image, show  # noqa: F401, I100


run_cli(cli_bootstrap)



import click

from kubetools import __version__
from kubetools.config import load_kubetools_config
from kubetools.log import setup_logging
from kubetools.settings import get_settings

from . import backends  # noqa


@click.group()
@click.option(
    '--env',
    envvar='KUBETOOLS_DEV_ENV',
    default=None,
    help='Override environment name.',
)
@click.option('--debug', is_flag=True)
@click.version_option(version=__version__, message='%(prog)s: v%(version)s')
@click.pass_context
def dev(ctx, env, debug=False):
    '''
    Kubetools dev client - develop apps with Docker.
    '''

    setup_logging(debug)

    settings = get_settings()

    if not env:
        env = settings.DEV_DEFAULT_ENV

    # Get the config and attach it to the context
    ctx.obj = load_kubetools_config(env=env, dev=True)



from .container import make_container_config
from .util import get_hash, make_dns_safe_name

DEPLOYMENT_REVISION_LIMIT = 5


def make_deployment_config(
    name, containers,
    replicas=1,
    labels=None,
    annotations=None,
    envvars=None,
    update_strategy=None,
):
    '''
    Builds a Kubernetes deployment configuration dict.
    '''

    labels = labels or {}
    annotations = annotations or {}

    # Build our container list
    kubernetes_containers = []
    for container_name, container in containers.items():
        kubernetes_containers.append(make_container_config(
            container_name, container,
            envvars=envvars,
            labels=labels,
            annotations=annotations,
        ))

    # The actual controller Kubernetes config
    controller = {
        'apiVersion': 'apps/v1',
        'kind': 'Deployment',
        'metadata': {
            'name': make_dns_safe_name(name),
            'labels': labels,
            'annotations': annotations,
        },
        'spec': {
            'revisionHistoryLimit': DEPLOYMENT_REVISION_LIMIT,
            'selector': {
                'matchLabels': labels,
            },
            'replicas': replicas,
            'template': {
                'metadata': {
                    'labels': labels,
                },
                'spec': {
                    'containers': kubernetes_containers,
                },
            },
        },
    }

    if update_strategy:
        controller['spec']['strategy'] = update_strategy

    container_volumes = {}
    for container in containers.values():
        if 'volumes' in container:
            for volume in container['volumes']:
                name = get_hash(volume)
                container_volumes[name] = {
                    'name': name,
                    'hostPath': {
                        'path': volume.split(':')[0],
                    },
                }

    if container_volumes:  # kube does not like an empty list
        controller['spec']['template']['spec']['volumes'] = container_volumes.values()

    for container in containers:
        if 'volumes' in container:
            for volume in container['volumes']:
                controller['spec']['template'].setdefault('volumes', []).append({
                    'name': get_hash(volume),
                    'hostPath': {
                        'path': volume.split(':')[0],
                    },
                })

    return controller



from hashlib import sha1


def make_dns_safe_name(name):
    return name.replace('_', '-')


def get_hash(name):
    if isinstance(name, str):
        name = name.encode()
    return sha1(name).hexdigest()[:6]


def copy_and_update(base, *extras):
    if base:
        new_base = base.copy()
    else:
        new_base = {}

    for extra in extras:
        if extra:
            new_base.update(extra)
    return new_base



from contextlib import contextmanager

import click


class Build(object):
    '''
    Build is a stub class that encapsulates the context and namespace for
    a given build, as well as accepting log entries.

    The kubetools server provides it's own build class, which also handles things
    like aborting builds via Redis and keeps them saved in the database.
    '''

    in_stage = False

    def __init__(self, env, namespace):
        self.env = env
        self.namespace = namespace

    def log_info(self, text, extra_detail=None, formatter=lambda s: s):
        '''
        Create BuildLog information.
        '''

        if extra_detail:
            text = f'{text}@{extra_detail}'

        if self.in_stage:
            text = f'    {text}'

        if formatter:
            text = formatter(text)

        click.echo(text)

    def log_warning(self, *args, **kwargs):
        kwargs['formatter'] = lambda s: click.style(s, 'yellow')
        self.log_info(*args, **kwargs)

    def log_error(self, *args, **kwargs):
        kwargs['formatter'] = lambda s: click.style(s, 'red')
        self.log_info(*args, **kwargs)

    @contextmanager
    def stage(self, stage_name):
        click.echo(f'--> {stage_name}')
        old_in_stage = self.in_stage
        self.in_stage = True
        yield
        self.in_stage = old_in_stage
        click.echo()
